@@ -11,11 +11,11 @@
 /*                                                                     */
 /***********************************************************************/
 
-/* $Id: signals.c,v 1.64 2002/05/06 12:02:26 xleroy Exp $ */
+/* $Id: signals.c,v 1.74 2003/07/17 15:11:03 xleroy Exp $ */
 
 #include <signal.h>
 #include <stdio.h>
-#if (defined(TARGET_sparc) && defined(SYS_solaris))
+#if defined(TARGET_sparc) && defined(SYS_solaris)
 #include <ucontext.h>
 #endif
 #include "alloc.h"
@@ -29,9 +29,15 @@
 #include "stack.h"
 #include "sys.h"
 #ifdef HAS_STACK_OVERFLOW_DETECTION
+#include <sys/time.h>
 #include <sys/resource.h>
 #endif
 
+extern char * code_area_start, * code_area_end;
+
+#define In_code_area(pc) \
+  ((char *)(pc) >= code_area_start && (char *)(pc) <= code_area_end)
+
 #ifdef _WIN32
 typedef void (*sighandler)(int sig);
 extern sighandler win32_signal(int sig, sighandler action);
@@ -39,10 +45,68 @@ extern sighandler win32_signal(int sig, sighandler action);
 #endif
 
 #if defined(TARGET_power) && defined(SYS_rhapsody)
-/* Confer machdep/ppc/unix_signal.c and mach/ppc/thread_status.h
-   in the Darwin sources */
-#define CONTEXT_GPR(ctx, regno) \
-  (((unsigned long *)((ctx)->sc_regs))[2 + (regno)])
+
+  #include <sys/utsname.h>
+
+  #define STRUCT_SIGCONTEXT void
+  #define CONTEXT_GPR(ctx, regno) (*context_gpr_p ((ctx), (regno)))
+  #define CONTEXT_PC(ctx) CONTEXT_GPR ((ctx), -2)
+  static int ctx_version = 0;
+  static void init_ctx (void)
+  {
+    struct utsname name;
+    if (uname (&name) == 0){
+      if (name.release[1] == '.' && name.release[0] <= '5'){
+        ctx_version = 1;
+      }else{
+        ctx_version = 2;
+      }
+    }else{
+      fatal_error ("cannot determine SIGCONTEXT format");
+    }
+  }
+
+  #ifdef DARWIN_VERSION_6
+    #include <sys/ucontext.h>
+    static unsigned long *context_gpr_p (void *ctx, int regno)
+    {
+      unsigned long *regs;
+      if (ctx_version == 0) init_ctx ();
+      if (ctx_version == 1){
+        /* old-style context (10.0 and 10.1) */
+        regs = (unsigned long *)(((struct sigcontext *)ctx)->sc_regs);
+      }else{
+        Assert (ctx_version == 2);
+        /* new-style context (10.2) */
+        regs = (unsigned long *)&(((struct ucontext *)ctx)->uc_mcontext->ss);
+      }
+      return &(regs[2 + regno]);
+    }
+  #else
+    #define SA_SIGINFO 0x0040
+    struct ucontext {
+      int       uc_onstack;
+      sigset_t  uc_sigmask;
+      struct sigaltstack uc_stack;
+      struct ucontext   *uc_link;
+      size_t    uc_mcsize;
+      unsigned long     *uc_mcontext;
+    };
+    static unsigned long *context_gpr_p (void *ctx, int regno)
+    {
+      unsigned long *regs;
+      if (ctx_version == 0) init_ctx ();
+      if (ctx_version == 1){
+        /* old-style context (10.0 and 10.1) */
+        regs = (unsigned long *)(((struct sigcontext *)ctx)->sc_regs);
+      }else{
+        Assert (ctx_version == 2);
+        /* new-style context (10.2) */
+        regs = (unsigned long *)((struct ucontext *)ctx)->uc_mcontext + 8;
+      }
+      return &(regs[2 + regno]);
+    }
+  #endif
 #endif
 
 #if defined(TARGET_power) && defined(SYS_aix)
@@ -154,6 +218,24 @@ void leave_blocking_section(void)
   async_signal_mode = 0;
 }
 
+#ifdef POSIX_SIGNALS
+static void reraise(int sig, int now)
+{
+  struct sigaction sa;
+  sa.sa_handler = 0;
+  sa.sa_flags = 0;
+  sigemptyset(&sa.sa_mask);
+  sigaction(sig, &sa, 0);
+  /* If the signal was sent using kill() (si_code == 0) or will
+     not recur then raise it here.  Otherwise return.  The
+     offending instruction will be reexecuted and the signal
+     will recur.  */
+  if (now == 1)
+    raise(sig);
+  return;
+}
+#endif
+
 #if defined(TARGET_alpha) || defined(TARGET_mips)
 void handle_signal(int sig, int code, struct sigcontext * context)
 #elif defined(TARGET_power) && defined(SYS_aix)
@@ -161,7 +243,11 @@ void handle_signal(int sig, int code, STRUCT_SIGCONTEXT * context)
 #elif defined(TARGET_power) && defined(SYS_elf)
 void handle_signal(int sig, struct sigcontext * context)
 #elif defined(TARGET_power) && defined(SYS_rhapsody)
+void handle_signal(int sig, int code, STRUCT_SIGCONTEXT * context)
+#elif defined(TARGET_power) && defined(SYS_bsd)
 void handle_signal(int sig, int code, struct sigcontext * context)
+#elif defined(TARGET_sparc) && defined(SYS_solaris)
+void handle_signal(int sig, int code, void * context)
 #else
 void handle_signal(int sig)
 #endif
@@ -182,30 +268,53 @@ void handle_signal(int sig)
     pending_signal = sig;
     young_limit = young_end;
     /* Some ports cache young_limit in a register.
-       Use the signal context to modify that register too, but not if
-       we are inside C code (i.e. caml_last_return_address != 0). */
-    if (caml_last_return_address == 0) {
+       Use the signal context to modify that register too, but only if
+       we are inside Caml code (not inside C code). */
 #if defined(TARGET_alpha)
+    if (In_code_area(context->sc_pc)) {
       /* Cached in register $14 */
       context->sc_regs[14] = (long) young_limit;
+    }
 #endif
 #if defined(TARGET_mips)
+    if (In_code_area(context->sc_pc)) {
       /* Cached in register $23 */
       context->sc_regs[23] = (int) young_limit;
+    }
 #endif
 #if defined(TARGET_power) && defined(SYS_aix)
+    if (caml_last_return_address == 0) {
       /* Cached in register 30 */
       CONTEXT_GPR(context, 30) = (ulong_t) young_limit;
+    }
 #endif
 #if defined(TARGET_power) && defined(SYS_elf)
+    if (caml_last_return_address == 0) {
       /* Cached in register 30 */
       context->regs->gpr[30] = (unsigned long) young_limit;
+    }
 #endif
 #if defined(TARGET_power) && defined(SYS_rhapsody)
+    if (In_code_area(CONTEXT_PC(context))) {
       /* Cached in register 30 */
       CONTEXT_GPR(context, 30) = (unsigned long) young_limit;
-#endif
     }
+#endif
+#if defined(TARGET_power) && defined(SYS_bsd)
+    if (caml_last_return_address == 0) {
+      /* Cached in register 30 */
+      context->sc_frame.fixreg[30] = (unsigned long) young_limit;
+    }
+#endif
+#if defined(TARGET_sparc) && defined(SYS_solaris)
+    { greg_t * gregs = ((ucontext_t *)context)->uc_mcontext.gregs;
+      if (In_code_area(gregs[REG_PC])) {
+      /* Cached in register l7, which is saved on the stack 7 words
+	 after the stack pointer.  */
+        ((long *)(gregs[REG_SP]))[7] = (long) young_limit;
+      }
+    }
+#endif
   }
 }
 
@@ -326,7 +435,11 @@ value install_signal_handler(value signal_number, value action) /* ML */
 #ifdef POSIX_SIGNALS
   sigact.sa_handler = act;
   sigemptyset(&sigact.sa_mask);
+#if defined(SYS_solaris) || defined(SYS_rhapsody)
+  sigact.sa_flags = SA_SIGINFO;
+#else
   sigact.sa_flags = 0;
+#endif
   if (sigaction(sig, &sigact, &oldsigact) == -1) sys_error(NO_ARG);
   oldact = oldsigact.sa_handler;
 #else
@@ -376,10 +489,9 @@ static void trap_handler(int sig, int code,
 #endif
 
 #if defined(TARGET_sparc) && defined(SYS_solaris)
-static void trap_handler(int sig, siginfo_t * info, void * arg)
+static void trap_handler(int sig, siginfo_t * info, void * context)
 {
-  ucontext_t * context;
-  int * sp;
+  long * sp;
 
   if (info->si_code != ILL_ILLTRP) {
     fprintf(stderr, "Fatal error: illegal instruction, code 0x%x\n",
@@ -387,8 +499,7 @@ static void trap_handler(int sig, siginfo_t * info, void * arg)
     exit(100);
   }
   /* Recover young_ptr and caml_exception_pointer from the %l5 and %l6 regs */
-  context = (ucontext_t *) arg;
-  sp = (int *) context->uc_mcontext.gregs[REG_SP];
+  sp = (long *) (((ucontext_t *)context)->uc_mcontext.gregs[REG_SP]);
   caml_exception_pointer = (char *) sp[5];
   young_ptr = (char *) sp[6];
   array_bound_error();
@@ -429,7 +540,7 @@ static void trap_handler(int sig, struct sigcontext * context)
 #endif
 
 #if defined(TARGET_power) && defined(SYS_rhapsody)
-static void trap_handler(int sig, int code, struct sigcontext * context)
+static void trap_handler(int sig, int code, STRUCT_SIGCONTEXT * context)
 {
   /* Unblock SIGTRAP */
   sigset_t mask;
@@ -443,6 +554,17 @@ static void trap_handler(int sig, int code, struct sigcontext * context)
 }
 #endif
 
+#if defined(TARGET_power) && defined(SYS_bsd)
+static void trap_handler(int sig, int code, struct sigcontext * context)
+{
+  /* Recover young_ptr and caml_exception_pointer from registers 31 and 29 */
+  caml_exception_pointer = (char *) context->sc_frame.fixreg[29];
+  young_ptr = (char *) context->sc_frame.fixreg[31];
+  array_bound_error();
+}
+#endif
+
+
 /* Machine- and OS-dependent handling of stack overflow */
 
 #ifdef HAS_STACK_OVERFLOW_DETECTION
@@ -522,7 +644,9 @@ void init_signals(void)
     struct sigaction act;
     act.sa_handler = (void (*)(int)) trap_handler;
     sigemptyset(&act.sa_mask);
-#if defined(SYS_rhapsody) || defined(SYS_aix)
+#if defined (SYS_rhapsody)
+    act.sa_flags = SA_SIGINFO;
+#elif defined (SYS_aix)
     act.sa_flags = 0;
 #else
     act.sa_flags = SA_NODEFER;
