@@ -5,14 +5,19 @@
 (*                                                                     *)
 (*        Daniel de Rauglaudre, projet Cristal, INRIA Rocquencourt     *)
 (*                                                                     *)
-(*  Copyright 2001 Institut National de Recherche en Informatique et   *)
+(*  Copyright 2002 Institut National de Recherche en Informatique et   *)
 (*  Automatique.  Distributed only by permission.                      *)
 (*                                                                     *)
 (***********************************************************************)
 
-(* $Id: mLast.mli,v 1.10 2002/07/19 14:53:44 mauny Exp $ *)
+(* $Id: mLast.mli,v 1.15 2003/07/16 12:50:07 mauny Exp $ *)
 
-(* Module [MLast]: abstract syntax tree *)
+(* Module [MLast]: abstract syntax tree
+
+   This is undocumented because the AST is not supposed to be used
+   directly; the good usage is to use the quotations representing
+   these values in concrete syntax (see the Camlp4 documentation).
+   See also the file q_MLast.ml in Camlp4 sources. *)
 
 type loc = (int * int);
 
@@ -30,8 +35,8 @@ type ctyp =
   | TyOlb of loc and string and ctyp
   | TyPol of loc and list string and ctyp
   | TyQuo of loc and string
-  | TyRec of loc and list (loc * string * bool * ctyp)
-  | TySum of loc and list (loc * string * list ctyp)
+  | TyRec of loc and bool and list (loc * string * bool * ctyp)
+  | TySum of loc and bool and list (loc * string * list ctyp)
   | TyTup of loc and list ctyp
   | TyUid of loc and string
   | TyVrn of loc and list row_field and option (option (list string)) ]
@@ -57,10 +62,13 @@ type patt =
   | PaArr of loc and list patt
   | PaChr of loc and string
   | PaInt of loc and string
+  | PaInt32 of loc and string
+  | PaInt64 of loc and string
+  | PaNativeInt of loc and string
   | PaFlo of loc and string
-  | PaLab of loc and string and patt
+  | PaLab of loc and string and option patt
   | PaLid of loc and string
-  | PaOlb of loc and string and patt and option expr
+  | PaOlb of loc and string and option (patt * option expr)
   | PaOrp of loc and patt and patt
   | PaRng of loc and patt and patt
   | PaRec of loc and list (patt * patt)
@@ -76,7 +84,9 @@ and expr =
   | ExApp of loc and expr and expr
   | ExAre of loc and expr and expr
   | ExArr of loc and list expr
-  | ExAss of loc and expr and expr
+  | ExAsf of loc                       (* assert False *)
+  | ExAsr of loc and expr              (* assert *)
+  | ExAss of loc and expr and expr     (* assignment *)
   | ExChr of loc and string
   | ExCoe of loc and expr and option ctyp and ctyp
   | ExFlo of loc and string
@@ -84,14 +94,17 @@ and expr =
   | ExFun of loc and list (patt * option expr * expr)
   | ExIfe of loc and expr and expr and expr
   | ExInt of loc and string
-  | ExLab of loc and string and expr
+  | ExInt32 of loc and string
+  | ExInt64 of loc and string
+  | ExNativeInt of loc and string
+  | ExLab of loc and string and option expr
   | ExLaz of loc and expr
   | ExLet of loc and bool and list (patt * expr) and expr
   | ExLid of loc and string
   | ExLmd of loc and string and module_expr and expr
   | ExMat of loc and expr and list (patt * option expr * expr)
   | ExNew of loc and list string
-  | ExOlb of loc and string and expr
+  | ExOlb of loc and string and option expr
   | ExOvr of loc and list (string * expr)
   | ExRec of loc and list (patt * expr) and option expr
   | ExSeq of loc and list expr
@@ -122,9 +135,11 @@ and sig_item =
   | SgExt of loc and string and ctyp and list string
   | SgInc of loc and module_type
   | SgMod of loc and string and module_type
+  | SgRecMod of loc and list (string * module_type)
   | SgMty of loc and string and module_type
   | SgOpn of loc and list string
   | SgTyp of loc and list type_decl
+  | SgUse of loc and string and list (sig_item * loc)
   | SgVal of loc and string and ctyp ]
 and with_constr =
   [ WcTyp of loc and list string and list (string * (bool * bool)) and ctyp
@@ -146,9 +161,11 @@ and str_item =
   | StExt of loc and string and ctyp and list string
   | StInc of loc and module_expr
   | StMod of loc and string and module_expr
+  | StRecMod of loc and list (string * module_type * module_expr)
   | StMty of loc and string and module_type
   | StOpn of loc and list string
   | StTyp of loc and list type_decl
+  | StUse of loc and string and list (str_item * loc)
   | StVal of loc and bool and list (patt * expr) ]
 and type_decl =
   ((loc * string) * list (string * (bool * bool)) * ctyp * list (ctyp * ctyp))
