@@ -10,11 +10,12 @@
 (*                                                                     *)
 (***********************************************************************)
 
-(* $Id: pa_o.ml,v 1.43 2002/07/19 14:53:45 mauny Exp $ *)
+(* $Id: pa_o.ml,v 1.52 2003/09/25 12:05:05 mauny Exp $ *)
 
 open Stdpp;
 open Pcaml;
 
+Pcaml.syntax_name.val := "OCaml";
 Pcaml.no_constructors_arity.val := True;
 
 do {
@@ -34,6 +35,7 @@ do {
   Grammar.Unsafe.clear_entry patt;
   Grammar.Unsafe.clear_entry ctyp;
   Grammar.Unsafe.clear_entry let_binding;
+  Grammar.Unsafe.clear_entry type_declaration;
   Grammar.Unsafe.clear_entry class_type;
   Grammar.Unsafe.clear_entry class_expr;
   Grammar.Unsafe.clear_entry class_sig_item;
@@ -50,11 +52,17 @@ value o2b =
 ;
 
 value mkumin loc f arg =
-  match arg with
-  [ <:expr< $int:n$ >> when int_of_string n > 0 ->
+  match (f, arg) with
+  [ ("-", <:expr< $int:n$ >>) when int_of_string n > 0 ->
       let n = "-" ^ n in
       <:expr< $int:n$ >>
-  | <:expr< $flo:n$ >> when float_of_string n > 0.0 ->
+  | ("-", MLast.ExInt32 loc n) when (Int32.of_string n) > 0l ->
+      MLast.ExInt32 loc ("-" ^ n)
+  | ("-", MLast.ExInt64 loc n) when (Int64.of_string n) > 0L ->
+      MLast.ExInt64 loc ("-" ^ n)
+  | ("-", MLast.ExNativeInt loc n) when (Nativeint.of_string n) > 0n ->
+      MLast.ExNativeInt loc ("-" ^ n)
+  | (_, <:expr< $flo:n$ >>) when float_of_string n > 0.0 ->
       let n = "-" ^ n in
       <:expr< $flo:n$ >>
   | _ ->
@@ -282,19 +290,43 @@ value test_typevar_list_dot =
 
 value constr_arity = ref [("Some", 1); ("Match_Failure", 1)];
 
-value rec constr_expr_arity =
+value rec is_expr_constr_call =
+  fun
+  [ <:expr< $uid:_$ >> -> True
+  | <:expr< $uid:_$.$e$ >> -> is_expr_constr_call e
+  | <:expr< $e$ $_$ >> -> is_expr_constr_call e
+  | _ -> False ]
+;
+
+value rec constr_expr_arity loc =
   fun
   [ <:expr< $uid:c$ >> ->
       try List.assoc c constr_arity.val with [ Not_found -> 0 ]
-  | <:expr< $uid:_$.$e$ >> -> constr_expr_arity e
+  | <:expr< $uid:_$.$e$ >> -> constr_expr_arity loc e
+  | <:expr< $e$ $_$ >> ->
+      if is_expr_constr_call e then
+        Stdpp.raise_with_loc loc (Stream.Error "currified constructor")
+      else 1
   | _ -> 1 ]
 ;
 
-value rec constr_patt_arity =
+value rec is_patt_constr_call =
+  fun
+  [ <:patt< $uid:_$ >> -> True
+  | <:patt< $uid:_$.$p$ >> -> is_patt_constr_call p
+  | <:patt< $p$ $_$ >> -> is_patt_constr_call p
+  | _ -> False ]
+;
+
+value rec constr_patt_arity loc =
   fun
   [ <:patt< $uid:c$ >> ->
       try List.assoc c constr_arity.val with [ Not_found -> 0 ]
-  | <:patt< $uid:_$.$p$ >> -> constr_patt_arity p
+  | <:patt< $uid:_$.$p$ >> -> constr_patt_arity loc p
+  | <:patt< $p$ $_$ >> ->
+      if is_patt_constr_call p then
+        Stdpp.raise_with_loc loc (Stream.Error "currified constructor")
+      else 1
   | _ -> 1 ]
 ;
 
@@ -322,11 +354,13 @@ value choose_tvar tpl =
 
 value rec patt_lid =
   fun
-  [ <:patt< $lid:i$ $p$ >> -> Some (i, [p])
-  | <:patt< $p1$ $p2$ >> ->
-      match patt_lid p1 with
-      [ Some (i, pl) -> Some (i, [p2 :: pl])
-      | None -> None ]
+  [ <:patt< $p1$ $p2$ >> ->
+      match p1 with
+      [ <:patt< $lid:i$ >> -> Some (MLast.loc_of_patt p1, i, [p2])
+      | _ ->
+          match patt_lid p1 with
+          [ Some (loc, i, pl) -> Some (loc, i, [p2 :: pl])
+          | None -> None ] ]
   | _ -> None ]
 ;
 
@@ -375,7 +409,7 @@ Pcaml.sync.val := sync;
 
 EXTEND
   GLOBAL: sig_item str_item ctyp patt expr module_type module_expr class_type
-    class_expr class_sig_item class_str_item let_binding;
+    class_expr class_sig_item class_str_item let_binding type_declaration;
   module_expr:
     [ [ "functor"; "("; i = UIDENT; ":"; t = module_type; ")"; "->";
         me = SELF ->
@@ -405,6 +439,8 @@ EXTEND
       | "include"; me = module_expr -> <:str_item< include $me$ >>
       | "module"; i = UIDENT; mb = module_binding ->
           <:str_item< module $i$ = $mb$ >>
+      | "module"; "rec"; nmtmes = LIST1 module_rec_binding SEP "and" ->
+          MLast.StRecMod loc nmtmes
       | "module"; "type"; i = UIDENT; "="; mt = module_type ->
           <:str_item< module type $i$ = $mt$ >>
       | "open"; i = mod_ident -> <:str_item< open $i$ >>
@@ -412,12 +448,12 @@ EXTEND
           <:str_item< type $list:tdl$ >>
       | "let"; r = OPT "rec"; l = LIST1 let_binding SEP "and"; "in";
         x = expr ->
-          let e = <:expr< let $rec:o2b r$ $list:l$ in $x$ >> in
+          let e = <:expr< let $opt:o2b r$ $list:l$ in $x$ >> in
           <:str_item< $exp:e$ >>
       | "let"; r = OPT "rec"; l = LIST1 let_binding SEP "and" ->
           match l with
           [ [(<:patt< _ >>, e)] -> <:str_item< $exp:e$ >>
-          | _ -> <:str_item< value $rec:o2b r$ $list:l$ >> ]
+          | _ -> <:str_item< value $opt:o2b r$ $list:l$ >> ]
       | "let"; "module"; m = UIDENT; mb = module_binding; "in"; e = expr ->
           <:str_item< let module $m$ = $mb$ in $e$ >>
       | e = expr -> <:str_item< $exp:e$ >> ] ]
@@ -434,6 +470,10 @@ EXTEND
           <:module_expr< ( $me$ : $mt$ ) >>
       | "="; me = module_expr -> <:module_expr< $me$ >> ] ]
   ;
+  module_rec_binding:
+    [ [ m = UIDENT; ":"; mt = module_type; "="; me = module_expr ->
+          (m, mt, me) ] ]
+  ;
   (* Module types *)
   module_type:
     [ [ "functor"; "("; i = UIDENT; ":"; t = SELF; ")"; "->"; mt = SELF ->
@@ -464,6 +504,8 @@ EXTEND
       | "include"; mt = module_type -> <:sig_item< include $mt$ >>
       | "module"; i = UIDENT; mt = module_declaration ->
           <:sig_item< module $i$ : $mt$ >>
+      | "module"; "rec"; mds = LIST1 module_rec_declaration SEP "and" ->
+          MLast.SgRecMod loc mds
       | "module"; "type"; i = UIDENT; "="; mt = module_type ->
           <:sig_item< module type $i$ = $mt$ >>
       | "module"; "type"; i = UIDENT ->
@@ -481,6 +523,9 @@ EXTEND
       | "("; i = UIDENT; ":"; t = module_type; ")"; mt = SELF ->
           <:module_type< functor ( $i$ : $t$ ) -> $mt$ >> ] ]
   ;
+  module_rec_declaration:
+    [ [ m = UIDENT; ":"; mt = module_type -> (m, mt)] ]
+  ;
   (* "with" constraints (additional type equations over signature
      components) *)
   with_constr:
@@ -498,7 +543,7 @@ EXTEND
     | "expr1"
       [ "let"; o = OPT "rec"; l = LIST1 let_binding SEP "and"; "in";
         x = expr LEVEL "top" ->
-          <:expr< let $rec:o2b o$ $list:l$ in $x$ >>
+          <:expr< let $opt:o2b o$ $list:l$ in $x$ >>
       | "let"; "module"; m = UIDENT; mb = module_binding; "in";
         e = expr LEVEL "top" ->
           <:expr< let module $m$ = $mb$ in $e$ >>
@@ -576,7 +621,7 @@ EXTEND
       | "-."; e = SELF -> <:expr< $mkumin loc "-." e$ >> ]
     | "apply" LEFTA
       [ e1 = SELF; e2 = SELF ->
-          match constr_expr_arity e1 with
+          match constr_expr_arity loc e1 with
           [ 1 -> <:expr< $e1$ $e2$ >>
           | _ ->
               match e2 with
@@ -584,15 +629,9 @@ EXTEND
                   List.fold_left (fun e1 e2 -> <:expr< $e1$ $e2$ >>) e1 el
               | _ -> <:expr< $e1$ $e2$ >> ] ]
       | "assert"; e = SELF ->
-          let f = <:expr< $str:input_file.val$ >> in
-          let bp = <:expr< $int:string_of_int (fst loc)$ >> in
-          let ep = <:expr< $int:string_of_int (snd loc)$ >> in
-          let raiser = <:expr< raise (Assert_failure ($f$, $bp$, $ep$)) >> in
           match e with
-          [ <:expr< False >> -> raiser
-          | _ ->
-              if no_assert.val then <:expr< () >>
-              else <:expr< if $e$ then () else $raiser$ >> ]
+          [ <:expr< False >> -> <:expr< assert False >>
+          | _ -> <:expr< assert ($e$) >> ]
       | "lazy"; e = SELF ->
           <:expr< lazy ($e$) >> ]
     | "." LEFTA
@@ -607,6 +646,9 @@ EXTEND
       | f = prefixop; e = SELF -> <:expr< $lid:f$ $e$ >> ]
     | "simple" LEFTA
       [ s = INT -> <:expr< $int:s$ >>
+      | s = INT32 -> MLast.ExInt32 loc s
+      | s = INT64 -> MLast.ExInt64 loc s
+      | s = NATIVEINT -> MLast.ExNativeInt loc s
       | s = FLOAT -> <:expr< $flo:s$ >>
       | s = STRING -> <:expr< $str:s$ >>
       | c = CHAR -> <:expr< $chr:c$ >>
@@ -628,6 +670,7 @@ EXTEND
       | "("; e = SELF; ":"; t = ctyp; ")" -> <:expr< ($e$ : $t$) >>
       | "("; e = SELF; ")" -> <:expr< $e$ >>
       | "begin"; e = SELF; "end" -> <:expr< $e$ >>
+      | "begin"; "end" -> <:expr< () >>
       | x = LOCATE ->
           let x =
             try
@@ -652,7 +695,7 @@ EXTEND
   let_binding:
     [ [ p = patt; e = fun_binding ->
           match patt_lid p with
-          [ Some (i, pl) ->
+          [ Some (loc, i, pl) ->
               let e =
                 List.fold_left (fun e p -> <:expr< fun $p$ -> $e$ >>) e pl
               in
@@ -715,7 +758,7 @@ EXTEND
       [ p1 = SELF; "::"; p2 = SELF -> <:patt< [$p1$ :: $p2$] >> ]
     | LEFTA
       [ p1 = SELF; p2 = SELF ->
-          match constr_patt_arity p1 with
+          match constr_patt_arity loc p1 with
           [ 1 -> <:patt< $p1$ $p2$ >>
           | n ->
               let p2 =
@@ -738,7 +781,13 @@ EXTEND
       [ s = LIDENT -> <:patt< $lid:s$ >>
       | s = UIDENT -> <:patt< $uid:s$ >>
       | s = INT -> <:patt< $int:s$ >>
+      | s = INT32 -> MLast.PaInt32 loc s
+      | s = INT64 -> MLast.PaInt64 loc s
+      | s = NATIVEINT -> MLast.PaNativeInt loc s
       | "-"; s = INT -> <:patt< $int:"-" ^ s$ >>
+      | "-"; s = INT32 -> MLast.PaInt32 loc ("-" ^ s)
+      | "-"; s = INT64 -> MLast.PaInt64 loc ("-" ^ s)
+      | "-"; s = NATIVEINT -> MLast.PaNativeInt loc ("-" ^ s)
       | "-"; s = FLOAT -> <:patt< $flo:"-" ^ s$ >>
       | s = FLOAT -> <:patt< $flo:s$ >>
       | s = STRING -> <:patt< $str:s$ >>
@@ -813,15 +862,23 @@ EXTEND
     [ [ "constraint"; t1 = ctyp; "="; t2 = ctyp -> (t1, t2) ] ]
   ;
   type_kind:
-    [ [ test_constr_decl; OPT "|";
-        cdl = LIST1 constructor_declaration SEP "|" ->
-          <:ctyp< [ $list:cdl$ ] >>
+    [ [ "private"; "{"; ldl = label_declarations; "}" ->
+          <:ctyp< private { $list:ldl$ } >>
+      | "private"; OPT "|"; 
+        cdl = LIST1 constructor_declaration SEP "|" -> <:ctyp< private [ $list:cdl$ ] >>
+      | test_constr_decl; OPT "|";
+        cdl = LIST1 constructor_declaration SEP "|" -> <:ctyp< [ $list:cdl$ ] >>
       | t = ctyp -> <:ctyp< $t$ >>
+      | t = ctyp; "="; "private"; "{"; ldl = label_declarations; "}" ->
+          <:ctyp< $t$ == private { $list:ldl$ } >>
       | t = ctyp; "="; "{"; ldl = label_declarations; "}" ->
           <:ctyp< $t$ == { $list:ldl$ } >>
+      | t = ctyp; "="; "private"; OPT "|"; cdl = LIST1 constructor_declaration SEP "|" ->
+          <:ctyp< $t$ == private [ $list:cdl$ ] >>
       | t = ctyp; "="; OPT "|"; cdl = LIST1 constructor_declaration SEP "|" ->
           <:ctyp< $t$ == [ $list:cdl$ ] >>
-      | "{"; ldl = label_declarations; "}" -> <:ctyp< { $list:ldl$ } >> ] ]
+      | "{"; ldl = label_declarations; "}" ->
+          <:ctyp< { $list:ldl$ } >> ] ]
   ;
   type_parameters:
     [ [ -> (* empty *) []
@@ -920,7 +977,11 @@ EXTEND
   class_fun_def:
     [ [ p = patt LEVEL "simple"; "->"; ce = class_expr ->
           <:class_expr< fun $p$ -> $ce$ >>
+      | p = labeled_patt; "->"; ce = class_expr ->
+          <:class_expr< fun $p$ -> $ce$ >>
       | p = patt LEVEL "simple"; cfd = SELF ->
+          <:class_expr< fun $p$ -> $cfd$ >>
+      | p = labeled_patt; cfd = SELF ->
           <:class_expr< fun $p$ -> $cfd$ >> ] ]
   ;
   class_expr:
@@ -928,7 +989,7 @@ EXTEND
       [ "fun"; cfd = class_fun_def -> cfd
       | "let"; rf = OPT "rec"; lb = LIST1 let_binding SEP "and"; "in";
         ce = SELF ->
-          <:class_expr< let $rec:o2b rf$ $list:lb$ in $ce$ >> ]
+          <:class_expr< let $opt:o2b rf$ $list:lb$ in $ce$ >> ]
     | "apply" LEFTA
       [ ce = SELF; e = expr LEVEL "label" ->
           <:class_expr< $ce$ $e$ >> ]
@@ -954,9 +1015,9 @@ EXTEND
   ;
   class_str_item:
     [ [ "inherit"; ce = class_expr; pb = OPT [ "as"; i = LIDENT -> i ] ->
-          <:class_str_item< inherit $ce$ $as:pb$ >>
-      | "val"; (lab, mf, e) = cvalue ->
-          <:class_str_item< value $mut:mf$ $lab$ = $e$ >>
+          <:class_str_item< inherit $ce$ $opt:pb$ >>
+      | "val"; mf = OPT "mutable"; lab = label; e = cvalue_binding ->
+          <:class_str_item< value $opt:o2b mf$ $lab$ = $e$ >>
       | "method"; "private"; "virtual"; l = label; ":"; t = poly_type ->
           <:class_str_item< method virtual private $l$ : $t$ >>
       | "method"; "virtual"; "private"; l = label; ":"; t = poly_type ->
@@ -975,15 +1036,13 @@ EXTEND
           <:class_str_item< type $t1$ = $t2$ >>
       | "initializer"; se = expr -> <:class_str_item< initializer $se$ >> ] ]
   ;
-  cvalue:
-    [ [ mf = OPT "mutable"; l = label; "="; e = expr -> (l, o2b mf, e)
-      | mf = OPT "mutable"; l = label; ":"; t = ctyp; "="; e = expr ->
-          (l, o2b mf, <:expr< ($e$ : $t$) >>)
-      | mf = OPT "mutable"; l = label; ":"; t = ctyp; ":>"; t2 = ctyp; "=";
-        e = expr ->
-          (l, o2b mf, <:expr< ($e$ : $t$ :> $t2$) >>)
-      | mf = OPT "mutable"; l = label; ":>"; t = ctyp; "="; e = expr ->
-          (l, o2b mf, <:expr< ($e$ :> $t$) >>) ] ]
+  cvalue_binding:
+    [ [ "="; e = expr -> e
+      | ":"; t = ctyp; "="; e = expr -> <:expr< ($e$ : $t$) >>
+      | ":"; t = ctyp; ":>"; t2 = ctyp; "="; e = expr ->
+          <:expr< ($e$ : $t$ :> $t2$) >>
+      | ":>"; t = ctyp; "="; e = expr ->
+          <:expr< ($e$ :> $t$) >> ] ]
   ;
   label:
     [ [ i = LIDENT -> i ] ]
@@ -1008,7 +1067,7 @@ EXTEND
   class_sig_item:
     [ [ "inherit"; cs = class_signature -> <:class_sig_item< inherit $cs$ >>
       | "val"; mf = OPT "mutable"; l = label; ":"; t = ctyp ->
-          <:class_sig_item< value $mut:o2b mf$ $l$ : $t$ >>
+          <:class_sig_item< value $opt:o2b mf$ $l$ : $t$ >>
       | "method"; "private"; "virtual"; l = label; ":"; t = poly_type ->
           <:class_sig_item< method virtual private $l$ : $t$ >>
       | "method"; "virtual"; "private"; l = label; ":"; t = poly_type ->
@@ -1035,7 +1094,7 @@ EXTEND
            MLast.ciNam = n; MLast.ciExp = cs} ] ]
   ;
   (* Expressions *)
-  expr: LEVEL "apply"
+  expr: LEVEL "simple"
     [ LEFTA
       [ "new"; i = class_longident -> <:expr< new $list:i$ >> ] ]
   ;
@@ -1058,7 +1117,7 @@ EXTEND
   (* Core types *)
   ctyp: LEVEL "simple"
     [ [ "#"; id = class_longident -> <:ctyp< # $list:id$ >>
-      | "<"; (ml, v) = meth_list; ">" -> <:ctyp< < $list:ml$ $v$ > >>
+      | "<"; (ml, v) = meth_list; ">" -> <:ctyp< < $list:ml$ $opt:v$ > >>
       | "<"; ">" -> <:ctyp< < > >> ] ]
   ;
   meth_list:
@@ -1091,12 +1150,19 @@ EXTEND
   (* Labels *)
   ctyp: AFTER "arrow"
     [ NONA
-      [ i = lident_colon; t = SELF -> <:ctyp< ~ $i$ : $t$ >>
-      | i = QUESTIONIDENT; ":"; t = SELF -> <:ctyp< ? $i$ : $t$ >> ] ]
+      [ i = lident_colon; t1 = ctyp LEVEL "star"; "->"; t2 = SELF ->
+          <:ctyp< ~ $i$ : $t1$ -> $t2$ >>
+      | i = OPTLABEL; t1 = ctyp LEVEL "star"; "->"; t2 = SELF ->
+          <:ctyp< ? $i$ : $t1$ -> $t2$ >>
+      | i = QUESTIONIDENT; ":"; t1 = ctyp LEVEL "star"; "->"; t2 = SELF ->
+          <:ctyp< ? $i$ : $t1$ -> $t2$ >>
+      | "?"; i=lident_colon;t1 = ctyp LEVEL "star"; "->"; t2 = SELF ->
+          <:ctyp< ? $i$ : $t1$ -> $t2$ >> ] ]
   ;
   ctyp: LEVEL "simple"
-    [ [ "["; OPT "|"; rfl = LIST0 row_field SEP "|"; "]" ->
+    [ [ "["; OPT "|"; rfl = LIST1 row_field SEP "|"; "]" ->
           <:ctyp< [ = $list:rfl$ ] >>
+      | "["; ">"; "]" -> <:ctyp< [ > $list:[]$ ] >>
       | "["; ">"; OPT "|"; rfl = LIST1 row_field SEP "|"; "]" ->
           <:ctyp< [ > $list:rfl$ ] >>
       | "[<"; OPT "|"; rfl = LIST1 row_field SEP "|"; "]" ->
@@ -1119,10 +1185,12 @@ EXTEND
   ;
   expr: AFTER "apply"
     [ "label"
-      [ i = TILDEIDENT; ":"; e = SELF -> <:expr< ~ $i$ : $e$ >>
+      [ i = LABEL; e = SELF -> <:expr< ~ $i$ : $e$ >>
       | i = TILDEIDENT -> <:expr< ~ $i$ >>
-      | i = QUESTIONIDENT; ":"; e = SELF -> <:expr< ? $i$ : $e$ >>
-      | i = QUESTIONIDENT -> <:expr< ? $i$ >> ] ]
+      | "~"; i = LIDENT -> <:expr< ~ $i$ >>
+      | i = OPTLABEL; e = SELF -> <:expr< ? $i$ : $e$ >>
+      | i = QUESTIONIDENT -> <:expr< ? $i$ >> 
+      | "?"; i = LIDENT -> <:expr< ? $i$ >> ] ]
   ;
   expr: LEVEL "simple"
     [ [ "`"; s = ident -> <:expr< ` $s$ >> ] ]
@@ -1138,37 +1206,43 @@ EXTEND
       | "#"; t = mod_ident -> <:patt< # $list:t$ >> ] ]
   ;
   labeled_patt:
-    [ [ i = TILDEIDENT; ":"; p = patt LEVEL "simple" ->
+    [ [ i = LABEL; p = patt LEVEL "simple" ->
            <:patt< ~ $i$ : $p$ >>
       | i = TILDEIDENT ->
            <:patt< ~ $i$ >>
+      | "~"; i=LIDENT -> <:patt< ~ $i$ >>
       | "~"; "("; i = LIDENT; ")" ->
            <:patt< ~ $i$ >>
       | "~"; "("; i = LIDENT; ":"; t = ctyp; ")" ->
            <:patt< ~ $i$ : ($lid:i$ : $t$) >>
-      | i = QUESTIONIDENT; ":"; j = LIDENT ->
+      | i = OPTLABEL; j = LIDENT ->
            <:patt< ? $i$ : ($lid:j$) >>
-      | i = QUESTIONIDENT; ":"; "("; p = patt; "="; e = expr; ")" ->
+      | i = OPTLABEL; "("; p = patt; "="; e = expr; ")" ->
           <:patt< ? $i$ : ( $p$ = $e$ ) >>
-      | i = QUESTIONIDENT; ":"; "("; p = patt; ":"; t = ctyp; ")" ->
+      | i = OPTLABEL; "("; p = patt; ":"; t = ctyp; ")" ->
           <:patt< ? $i$ : ( $p$ : $t$ ) >>
-      | i = QUESTIONIDENT; ":"; "("; p = patt; ":"; t = ctyp; "=";
+      | i = OPTLABEL; "("; p = patt; ":"; t = ctyp; "=";
         e = expr; ")" ->
           <:patt< ? $i$ : ( $p$ : $t$ = $e$ ) >>
-      | i = QUESTIONIDENT -> <:patt< ? $i$ : ($lid:i$) >>
+      | i = QUESTIONIDENT -> <:patt< ? $i$ >>
+      | "?"; i = LIDENT -> <:patt< ? $i$ >>
       | "?"; "("; i = LIDENT; "="; e = expr; ")" ->
-          <:patt< ? $i$ : ( $lid:i$ = $e$ ) >>
+          <:patt< ? ( $lid:i$ = $e$ ) >>
       | "?"; "("; i = LIDENT; ":"; t = ctyp; "="; e = expr; ")" ->
-          <:patt< ? $i$ : ( $lid:i$ : $t$ = $e$ ) >>
+          <:patt< ? ( $lid:i$ : $t$ = $e$ ) >>
       | "?"; "("; i = LIDENT; ")" ->
           <:patt< ? $i$ >>
       | "?"; "("; i = LIDENT; ":"; t = ctyp; ")" ->
-          <:patt< ? $i$ : ( $lid:i$ : $t$ ) >> ] ]
+          <:patt< ? ( $lid:i$ : $t$ ) >> ] ]
   ;
   class_type:
-    [ [ i = lident_colon; t = ctyp LEVEL "ctyp1"; "->"; ct = SELF ->
+    [ [ i = lident_colon; t = ctyp LEVEL "star"; "->"; ct = SELF ->
           <:class_type< [ ~ $i$ : $t$ ] -> $ct$ >>
-      | i = QUESTIONIDENT; ":"; t = ctyp LEVEL "ctyp1"; "->"; ct = SELF ->
+      | i = OPTLABEL; t = ctyp LEVEL "star"; "->"; ct = SELF ->
+          <:class_type< [ ? $i$ : $t$ ] -> $ct$ >> 
+      | i = QUESTIONIDENT; ":"; t = ctyp LEVEL "star"; "->"; ct = SELF ->
+          <:class_type< [ ? $i$ : $t$ ] -> $ct$ >> 
+      | "?"; i = LIDENT; ":"; t = ctyp LEVEL "star"; "->"; ct = SELF ->
           <:class_type< [ ? $i$ : $t$ ] -> $ct$ >> ] ]
   ;
   class_fun_binding:
@@ -1214,3 +1288,6 @@ EXTEND
       | "#"; n = LIDENT; dp = OPT expr -> <:str_item< # $n$ $opt:dp$ >> ] ]
   ;
 END;
+
+Pcaml.add_option "-no_quot" (Arg.Set Plexer.no_quotations)
+  "Don't parse quotations, allowing to use, e.g. \"<:>\" as token";
