@@ -1,21 +1,26 @@
-(*************************************************************************)
-(*                                                                       *)
-(*                Objective Caml LablTk library                          *)
-(*                                                                       *)
-(*         Jun Furuse, projet Cristal, INRIA Rocquencourt                *)
-(*                                                                       *)
-(*   Copyright 1999 Institut National de Recherche en Informatique et    *)
-(*   en Automatique and Kyoto University.  All rights reserved.          *)
-(*   This file is distributed under the terms of the GNU Library         *)
-(*   General Public License, with the special exception on linking       *)
-(*   described in file ../../../LICENSE.                                 *)
-(*                                                                       *)
-(*************************************************************************)
+(***********************************************************************)
+(*                                                                     *)
+(*                 MLTk, Tcl/Tk interface of Objective Caml            *)
+(*                                                                     *)
+(*    Francois Rouaix, Francois Pessaux, Jun Furuse and Pierre Weis    *)
+(*               projet Cristal, INRIA Rocquencourt                    *)
+(*            Jacques Garrigue, Kyoto University RIMS                  *)
+(*                                                                     *)
+(*  Copyright 2002 Institut National de Recherche en Informatique et   *)
+(*  en Automatique and Kyoto University.  All rights reserved.         *)
+(*  This file is distributed under the terms of the GNU Library        *)
+(*  General Public License, with the special exception on linking      *)
+(*  described in file LICENSE found in the Objective Caml source tree. *)
+(*                                                                     *)
+(***********************************************************************)
 
-(* $Id: fileselect.ml,v 1.17 2001/12/07 13:40:08 xleroy Exp $ *)
+(* $Id: fileselect.ml,v 1.19 2002/04/26 12:16:12 furuse Exp $ *)
 
 (* file selection box *)
 
+(* This file selecter works only under the OS with the full unix support.
+   For the portability, Tk.getOpenFile and Tk.getSaveFile are recommended. *)
+
 open StdLabels
 open UnixLabels
 open Str
@@ -70,25 +75,25 @@ let subshell cmd =
 (***************************************************************** Path name *)
 
 (* find directory name which doesn't contain "?*[" *)
-let dirget = regexp "^\([^\*?[]*/\)\(.*\)"
+let dirget = regexp "^\\([^\\*?[]*/\\)\\(.*\\)"
 
 let parse_filter src = 
   (* replace // by / *)
   let s = global_replace (regexp "/+") "/" src in
   (* replace /./ by / *)
-  let s = global_replace (regexp "/\./") "/" s in
+  let s = global_replace (regexp "/\\./") "/" s in
   (* replace ????/../ by "" *)
   let s = global_replace
-      (regexp "\([^/]\|[^\./][^/]\|[^/][^\./]\|[^/][^/]+\)/\.\./") 
+      (regexp "\\([^/]\\|[^\\./][^/]\\|[^/][^\\./]\\|[^/][^/]+\\)/\\.\\./") 
       ""
       s in
   (* replace ????/..$ by "" *)
   let s = global_replace
-      (regexp "\([^/]\|[^\./][^/]\|[^/][^\./]\|[^/][^/]+\)/\.\.$") 
+      (regexp "\\([^/]\\|[^\\./][^/]\\|[^/][^\\./]\\|[^/][^/]+\\)/\\.\\.$") 
       ""
       s in
   (* replace ^/../../ by / *)
-  let s = global_replace (regexp "^\(/\.\.\)+/") "/" s in
+  let s = global_replace (regexp "^\\(/\\.\\.\\)+/") "/" s in
   if string_match dirget s 0 then 
     let dirs = matched_group 1 s
     and ptrn = matched_group 2 s
