@@ -6,11 +6,12 @@
 /*                                                                     */
 /*  Copyright 1996 Institut National de Recherche en Informatique et   */
 /*  en Automatique.  All rights reserved.  This file is distributed    */
-/*  under the terms of the GNU Library General Public License.         */
+/*  under the terms of the GNU Library General Public License, with    */
+/*  the special exception on linking described in file ../../LICENSE.  */
 /*                                                                     */
 /***********************************************************************/
 
-/* $Id: getcwd.c,v 1.12 2000/04/05 18:30:19 doligez Exp $ */
+/* $Id: getcwd.c,v 1.14 2001/12/07 13:40:28 xleroy Exp $ */
 
 #include <mlvalues.h>
 #include <alloc.h>
@@ -30,7 +31,7 @@
 
 #ifdef HAS_GETCWD
 
-value unix_getcwd(value unit)     /* ML */
+CAMLprim value unix_getcwd(value unit)
 {
   char buff[PATH_MAX];
   if (getcwd(buff, sizeof(buff)) == 0) uerror("getcwd", Nothing);
@@ -40,7 +41,7 @@ value unix_getcwd(value unit)     /* ML */
 #else
 #ifdef HAS_GETWD
 
-value unix_getcwd(value unit)
+CAMLprim value unix_getcwd(value unit)
 {
   char buff[PATH_MAX];
   if (getwd(buff) == 0) uerror("getcwd", copy_string(buff));
@@ -49,7 +50,7 @@ value unix_getcwd(value unit)
 
 #else
 
-value unix_getcwd(value unit)
+CAMLprim value unix_getcwd(value unit)
 { invalid_argument("getcwd not implemented"); }
 
 #endif
