@@ -11,11 +11,7 @@
 /*                                                                     */
 /***********************************************************************/
 
-/* $Id: readdir.c,v 1.10 2001/12/07 13:40:32 xleroy Exp $ */
-
-/* This avoids "overflow" errors when reading from very large directories
-   (PR#609) */
-#define _FILE_OFFSET_BITS 64
+/* $Id: readdir.c,v 1.11 2003/03/03 17:12:33 xleroy Exp $ */
 
 #include <mlvalues.h>
 #include <fail.h>
