@@ -10,7 +10,7 @@
 (*                                                                     *)
 (***********************************************************************)
 
-(* $Id: genprintval.ml,v 1.31 2002/06/26 14:52:17 xleroy Exp $ *)
+(* $Id: genprintval.ml,v 1.36 2003/07/02 09:14:32 xleroy Exp $ *)
 
 (* To print values *)
 
@@ -83,7 +83,9 @@ module Make(O : OBJ)(EVP : EVALPATH with type value = O.t) = struct
     let outval_of_untyped_exception bucket =
       let name = (O.obj(O.field(O.field bucket 0) 0) : string) in
       let args =
-        if (name = "Match_failure" || name = "Assert_failure")
+        if (name = "Match_failure"
+            || name = "Assert_failure"
+            || name = "Undefined_recursive_module")
         && O.size bucket = 2
         && O.tag(O.field bucket 1) = 0
         then outval_of_untyped_exception_args (O.field bucket 1) 0
@@ -102,14 +104,11 @@ module Make(O : OBJ)(EVP : EVALPATH with type value = O.t) = struct
       Pident(Ident.create "print_string"), Predef.type_string,
         (fun x -> Oval_string (O.obj x : string));
       Pident(Ident.create "print_int32"), Predef.type_int32,
-        (fun x -> Oval_stuff ("<int32 " ^
-                              Int32.to_string (O.obj x : int32) ^ ">"));
+        (fun x -> Oval_int32 (O.obj x : int32));
       Pident(Ident.create "print_nativeint"), Predef.type_nativeint,
-        (fun x -> Oval_stuff ("<nativeint " ^
-                             Nativeint.to_string (O.obj x : nativeint) ^ ">"));
+        (fun x -> Oval_nativeint (O.obj x : nativeint));
       Pident(Ident.create "print_int64"), Predef.type_int64,
-        (fun x -> Oval_stuff ("<int64 " ^
-                              Int64.to_string (O.obj x : int64) ^ ">"))
+        (fun x -> Oval_int64 (O.obj x : int64))
     ] : (Path.t * type_expr * (O.t -> Outcometree.out_value)) list)
 
     let install_printer path ty fn =
@@ -243,7 +242,7 @@ module Make(O : OBJ)(EVP : EVALPATH with type value = O.t) = struct
                     tree_of_val depth obj
                       (try Ctype.apply env decl.type_params body ty_list with
                          Ctype.Cannot_apply -> abstract_type)
-                | {type_kind = Type_variant constr_list} ->
+                | {type_kind = Type_variant(constr_list, priv)} ->
                     let tag =
                       if O.is_block obj
                       then Cstr_block(O.tag obj)
@@ -258,8 +257,8 @@ module Make(O : OBJ)(EVP : EVALPATH with type value = O.t) = struct
                         constr_args in
                     tree_of_constr_with_args (tree_of_constr env path)
                                            constr_name 0 depth obj ty_args
-                | {type_kind = Type_record(lbl_list, rep)} ->
-                    match check_depth depth obj ty with
+                | {type_kind = Type_record(lbl_list, rep, priv)} ->
+                    begin match check_depth depth obj ty with
                       Some x -> x
                     | None ->
                         let rec tree_of_fields pos = function
@@ -279,6 +278,7 @@ module Make(O : OBJ)(EVP : EVALPATH with type value = O.t) = struct
                               (lid, v) :: tree_of_fields (pos + 1) remainder
                         in
                         Oval_record (tree_of_fields 0 lbl_list)
+                    end
               with
                 Not_found ->                (* raised by Env.find_type *)
                   Oval_stuff "<abstr>"
