@@ -10,7 +10,7 @@
 (*                                                                     *)
 (***********************************************************************)
 
-(* $Id: toploop.ml,v 1.74 2002/07/25 15:45:14 xleroy Exp $ *)
+(* $Id: toploop.ml,v 1.83 2003/08/21 13:55:30 xleroy Exp $ *)
 
 (* The interactive toplevel loop *)
 
@@ -104,6 +104,10 @@ let print_location = Location.print
 let print_warning = Location.print_warning
 let input_name = Location.input_name
 
+(* Hooks for initialization *)
+
+let toplevel_startup_hook = ref (fun () -> ())
+
 (* Load in-core and execute a lambda term *)
 
 let may_trace = ref false (* Global lock on tracing *)
@@ -122,6 +126,7 @@ let load_lambda ppf lam =
   let can_free = (fun_code = []) in
   let initial_symtable = Symtable.current_state() in
   Symtable.patch_object code reloc;
+  Symtable.check_global_initialized reloc;
   Symtable.update_global_table();
   try
     may_trace := true;
@@ -277,6 +282,7 @@ let use_file ppf name =
     let filename = find_in_path !Config.load_path name in
     let ic = open_in_bin filename in
     let lb = Lexing.from_channel ic in
+    Location.init lb filename;
     (* Skip initial #! line if any *)
     Lexer.skip_sharp_bang lb;
     let success =
@@ -330,24 +336,25 @@ let refill_lexbuf buffer len =
     | Exit -> !i
   end
 
-(* Discard data left in lexer buffer. *)
-
-let empty_lexbuf lb =
-  lb.lex_curr_pos <- 0;
-  lb.lex_abs_pos <- 0;
-  lb.lex_buffer_len <- 0
-
 (* Toplevel initialization. Performed here instead of at the
    beginning of loop() so that user code linked in with ocamlmktop
    can call directives from Topdirs. *)
 
 let _ =
   Sys.interactive := true;
-  Symtable.init_toplevel();
-  Compile.init_path()
+  let crc_intfs = Symtable.init_toplevel() in
+  Compile.init_path();
+  List.iter
+    (fun (name, crc) ->
+      Consistbl.set Env.crc_units name crc Sys.executable_name)
+    crc_intfs
 
 let load_ocamlinit ppf =
+  let home_init = 
+    try Filename.concat (Sys.getenv "HOME") ".ocamlinit"
+    with Not_found -> ".ocamlinit" in
   if Sys.file_exists ".ocamlinit" then ignore(use_silently ppf ".ocamlinit")
+  else if Sys.file_exists home_init then ignore(use_silently ppf home_init)
 
 let set_paths () =
   (* Add whatever -I options have been specified on the command line,
@@ -373,8 +380,9 @@ let loop ppf =
   Sys.catch_break true;
   load_ocamlinit ppf;
   while true do
+    let snap = Btype.snapshot () in
     try
-      empty_lexbuf lb;
+      Lexing.flush_input lb;
       Location.reset();
       first_line := true;
       let phr = try !parse_toplevel_phrase lb with Exit -> raise PPerror in
@@ -382,22 +390,17 @@ let loop ppf =
       ignore(execute_phrase true ppf phr)
     with
     | End_of_file -> exit 0
-    | Sys.Break -> fprintf ppf "Interrupted.@."
+    | Sys.Break -> fprintf ppf "Interrupted.@."; Btype.backtrack snap
     | PPerror -> ()
-    | x -> Errors.report_error ppf x
+    | x -> Errors.report_error ppf x; Btype.backtrack snap
   done
 
 (* Execute a script *)
 
 let run_script ppf name args =
-  let rec find n =
-    if n >= Array.length args then invalid_arg "Toploop.run_script";
-    if args.(n) = name then n else find (n+1) 
-  in
-  let pos = find 0 in
-  let len = Array.length args - pos in
+  let len = Array.length args in
   if Array.length Sys.argv < len then invalid_arg "Toploop.run_script";
-  Array.blit args pos Sys.argv 0 len;
+  Array.blit args 0 Sys.argv 0 len;
   Obj.truncate (Obj.repr Sys.argv) len;
   Arg.current := 0;
   Compile.init_path();
