@@ -10,7 +10,7 @@
 (*                                                                     *)
 (***********************************************************************)
 
-(* $Id: printtyp.mli,v 1.21 2001/10/31 15:41:11 guesdon Exp $ *)
+(* $Id: printtyp.mli,v 1.24 2002/04/18 07:27:45 garrigue Exp $ *)
 
 (* Printing functions *)
 
@@ -32,7 +32,7 @@ val type_scheme: formatter -> type_expr -> unit
 (* Maxence *)
 val reset_names: unit -> unit
 val type_scheme_max: ?b_reset_names: bool ->
-	formatter -> type_expr -> unit
+        formatter -> type_expr -> unit
 (* Fin Maxence *)
 val tree_of_value_description: Ident.t -> value_description -> out_sig_item
 val value_description: Ident.t -> formatter -> value_description -> unit
@@ -64,7 +64,3 @@ val report_unification_error:
 val report_subtyping_error:
     formatter -> (type_expr * type_expr) list ->
     string -> (type_expr * type_expr) list -> unit
-
-val outcome_type: (formatter -> out_type -> unit) ref
-val outcome_sig_item: (formatter -> out_sig_item -> unit) ref
-val outcome_signature: (formatter -> out_sig_item list -> unit) ref
