@@ -10,13 +10,14 @@
 (*                                                                     *)
 (***********************************************************************)
 
-(* $Id: warnings.ml,v 1.14 2002/05/24 14:34:37 maranget Exp $ *)
+(* $Id: warnings.ml,v 1.17 2003/05/02 12:52:11 weis Exp $ *)
 
 (* Please keep them in alphabetical order *)
 
 type t =                             (* A is all *)
   | Comment of string                (* C *)
   | Deprecated                       (* D *)
+  | Fragile_pat of string            (* E *)
   | Partial_application              (* F *)
   | Labels_omitted                   (* L *)
   | Method_override of string list   (* M *)
@@ -31,6 +32,7 @@ type t =                             (* A is all *)
 let letter = function        (* 'a' is all *)
   | Comment _ ->                'c'
   | Deprecated ->               'd'
+  | Fragile_pat _ ->              'e'
   | Partial_application ->      'f'
   | Labels_omitted ->           'l'
   | Method_override _ ->        'm'
@@ -42,7 +44,7 @@ let letter = function        (* 'a' is all *)
 ;;
 
 let check c =
-  try ignore (String.index "acdflmpsuvxACDFLMPSUVX" c)
+  try ignore (String.index "acdeflmpsuvxACDEFLMPSUVX" c)
   with _ -> raise (Arg.Bad (Printf.sprintf "unknown warning option %c" c))
 ;;    
 
@@ -79,7 +81,7 @@ let parse_options iserr s =
   done
 ;;
 
-let () = parse_options false "l";;
+let () = parse_options false "el";;
 
 let message = function
   | Partial_match "" -> "this pattern-matching is not exhaustive."
@@ -88,6 +90,13 @@ let message = function
        Here is an example of a value that is not matched:\n" ^ s
   | Unused_match -> "this match case is unused."
   | Unused_pat   -> "this pattern is unused."
+  | Fragile_pat "" ->
+      "this pattern is fragile. It would hide\n\
+       the addition of new constructors to the data types it matches."
+  | Fragile_pat s ->
+      "this pattern is fragile. It would hide\n\
+       the addition of new constructors to the data types it matches.\n\
+       Here is an example of a more robust pattern:\n" ^ s
   | Labels_omitted ->
       "labels were omitted in the application of this function."
   | Method_override slist ->
